/*
 * Copyright E. Bigeon (2014)
 *
 * emmanuel@bigeon.fr
 *
 * This software is a computer program whose purpose is to
 * Socket implementation of GCLC.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software.  You can  use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and,  more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms.
 */
package fr.bigeon.gclc.socket;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintWriter;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.SocketException;
import java.nio.charset.Charset;
import java.util.Arrays;
import java.util.logging.Level;
import java.util.logging.Logger;

import fr.bigeon.gclc.ConsoleApplication;
import fr.bigeon.gclc.manager.ConsoleManager;
import fr.bigeon.gclc.manager.PipedConsoleManager;
import fr.bigeon.gclc.manager.ReadingRunnable;
import fr.bigeon.smu.StringEncoder;

/** This is a socket communicating console consoleManager
 * <p>
 * To use this application, the following flow should be used:
 *
 * <pre>
 * SocketConsoleApplicationShell shell = new SocketConsoleApplicationShell();
 * ConsoleApplication myApplication = new MyConsoleApplication(shell.getConsoleManager(), ...);
 * shell.setApplication(myApplication);
 * Thread th = new Thread(shell);
 * th.start();
 * </pre>
 * <p>
 * This will start the application in a separate thread. The application will be
 * listening on the given socket and writing back on it. If this is all your
 * application, you should then {@link Thread#join()} the thread to wait for the
 * end of the execution.
 *
 * @author Emmanuel Bigeon */
public class SocketConsoleApplicationShell implements Runnable {

    /**
     * 
     */
    private static final String INTERRUPTION_WHILE_WORKING = "Interruption while application was working"; //$NON-NLS-1$
    /** The end of line character */
    protected static final String EOL = "\n"; //$NON-NLS-1$
    /** The encoder */
    private static final StringEncoder ENCODER = new StringEncoder("%", //$NON-NLS-1$
            Arrays.asList(EOL));
    /** The class logger */
    private static final Logger LOGGER = Logger
            .getLogger(SocketConsoleApplicationShell.class.getName());
    protected static final long ONE_TENTH_OF_SECOND = 100;
    private static final long ONE_SEC = 100;
    /** The listening port */
    private final int port;
    /** The input */
    private final PipedInputStream consoleInput = new PipedInputStream();
    /** The application */
    private ConsoleApplication app;
    /** The session closing command */
    private final String close;
    /** The running status */
    private boolean running;

//    /** The console manager implementation */
//    private final ThreadedServerConsoleManager consoleManager = new ThreadedServerConsoleManager(
//            ENCODER, promptingLock);
    /** The console manager implementation */
    private final PipedConsoleManager consoleManager;
    /** The auto close flag. if this is true, every request closes the session
     * after its call */
    private final boolean autoClose;
    /** The server socket */
    private ServerSocket serverSocket;
    /** The application shutdown string */
    private final String applicationShutdown;
    /** The charset for the communication. */
    private final Charset charset;

    /** Create a socket application shell which will listen on the given port
     * and close session upon the provided string reception by client
     *
     * @param port the port to listen to
     * @param close the session closing command
     * @param applicationShutdown the appication shut down command
     * @param charset the charset for communication
     * @throws IOException if the manager could not be created */
    public SocketConsoleApplicationShell(int port, String close,
            String applicationShutdown, Charset charset) throws IOException {
        this.port = port;
        this.close = close;
        this.applicationShutdown = applicationShutdown;
        this.autoClose = false;
        this.charset = charset;
        //
        consoleManager = new PipedConsoleManager();
    }

    /** Create a socket application shell which will listen on the given port
     * and auto close session after one instruction
     *
     * @param port the port to listen to
     * @param autoClose if the session must be closed once the request has been
     *            sent
     * @param applicationShutdown the appication shut down command
     * @param charset the charset for communication
     * @throws IOException if the manager could not be created */
    public SocketConsoleApplicationShell(int port, boolean autoClose,
            String applicationShutdown, Charset charset) throws IOException {
        this.port = port;
        this.autoClose = autoClose;
        this.applicationShutdown = applicationShutdown;
        this.close = autoClose ? null : "close"; //$NON-NLS-1$
        this.charset = charset;
        //
        consoleManager = new PipedConsoleManager();
    }

    /* (non-Javadoc)
     * @see java.lang.Runnable#run() */
    @Override
    public void run() {
        // Create the server
        try (ServerSocket actualServerSocket = new ServerSocket(port)) {
            this.serverSocket = actualServerSocket;
            running = true;
            // Create the streams
            try (PipedOutputStream outStream = new PipedOutputStream();
                 BufferedWriter writer = new BufferedWriter(
                         new OutputStreamWriter(outStream, charset));
                 InputStreamReader isr = new InputStreamReader(consoleInput,
                         charset);
                 BufferedReader inBuf = new BufferedReader(isr)) {
                consoleInput.connect(outStream);
//                consoleManager.setInput(inBuf);
                runSokectServer(writer);
                // Close the application
                // Pass command to application
                if (app.isRunning()) {
                    writer.write(applicationShutdown + EOL);
                    writer.flush();
                }
            }
        } catch (

        final IOException e) {
            LOGGER.log(Level.SEVERE,
                    "Communication error between client and server", e); //$NON-NLS-1$
        }
    }

    /** @param writer the writer to the application
     * @throws IOException if the communication with the client failed */
    private void runSokectServer(BufferedWriter writer) throws IOException {
        final ConsoleRunnable runnable = new ConsoleRunnable(app);
        Thread appThNext = new Thread(runnable, "gclc-ctrl");
        appThNext.start();
        while (running) {
            LOGGER.info("Opening client"); //$NON-NLS-1$
            try (Socket clientSocket = serverSocket.accept();
                 PrintWriter out = new PrintWriter(new OutputStreamWriter(
                         clientSocket.getOutputStream(), charset), true);
                 InputStreamReader isr = new InputStreamReader(
                         clientSocket.getInputStream(), charset);
                 BufferedReader in = new BufferedReader(isr);) {
                // this is not threaded to avoid several clients at the same
                // time

                // Initiate application
                if (!runnable.isApplicationRunning()) {
                    LOGGER.info("Start application");
                    runnable.restart();
                    synchronized (this) {
                        try {
                            wait(ONE_SEC);
                        } catch (InterruptedException e) {
                            LOGGER.log(Level.SEVERE,
                                    "Interruption in application start", e);
                        }
                    }
                } else {
                    LOGGER.info("Reconnect to application");
                    out.println("Reconnected"); //$NON-NLS-1$
                    out.println(consoleManager.getPrompt());
                }
                communicate(clientSocket, out, in);
            } catch (SocketException e) {
                LOGGER.log(Level.INFO, "Socket closed"); //$NON-NLS-1$
                LOGGER.log(Level.FINE,
                        "Socket closed with exception (probably due to server interruption)", //$NON-NLS-1$
                        e);
            }
            LOGGER.info("Closing client"); //$NON-NLS-1$
        }
        runnable.setRunning(false);
        consoleManager.type(applicationShutdown);
        LOGGER.info("Out client"); //$NON-NLS-1$
    }

    /** active communication between server and client
     * 
     * @param writer the writer to the application
     * @param in the input from the client
     * @throws IOException if the communication failed */
    private void communicate(final Socket socket, final PrintWriter writer,
                             BufferedReader in) throws IOException {
        Thread th = new Thread(new Runnable() {

            @SuppressWarnings("synthetic-access")
            @Override
            public void run() {
                try {
                    while (!socket.isOutputShutdown()) {
                        while (!socket.isOutputShutdown() &&
                               !consoleManager.available()) {
                            try {
                                synchronized (this) {
                                    wait(ONE_TENTH_OF_SECOND);
                                }
                            } catch (InterruptedException e) {
                                LOGGER.log(Level.SEVERE, "Interrupted wait", //$NON-NLS-1$
                                        e);
                                return;
                            }
                        }
                        if (socket.isOutputShutdown()) {
                            return;
                        }
                        String m = consoleManager.readNextLine();
                        writer.println(m);
                    }
                } catch (IOException e) {
                    LOGGER.log(Level.SEVERE, "Unexpected problem in manager", //$NON-NLS-1$
                            e);
                }
            }
        }, "ClientComm"); //$NON-NLS-1$
        th.start();
        if (autoClose) {
            communicateOnce(socket, in);
        } else {
            communicateLoop(socket, in);
        }
    }

    /** @param in the input from the client
     * @throws IOException if the communication failed */
    private void communicateOnce(Socket socket,
                                 BufferedReader in) throws IOException {
        ReadingRunnable reading = new ReadingRunnable(in);
        Thread th = new Thread(reading, "gclcToApp");
        th.start();
        String ln;
        if (app.isRunning()) {
            while (app.isRunning() && !reading.hasMessage()) {
                synchronized (this) {
                    try {
                        wait(ONE_SEC);
                    } catch (InterruptedException e) {
                        LOGGER.log(Level.SEVERE, "Wait interrupted", e);
                    }
                }
            }
            if (!app.isRunning()) {
                return;
            }
            ln = reading.getMessage();
            if (ln.equals(close)) {
                return;
            }
            // Pass command to application
            consoleManager.type(ln);
        }
        reading.setRunning(false);
        socket.shutdownOutput();
    }

    /** @param in the input from the client
     * @throws IOException if the communication failed */
    private void communicateLoop(Socket socket,
                                 BufferedReader in) throws IOException {
        ReadingRunnable reading = new ReadingRunnable(in);
        Thread th = new Thread(reading, "gclcToApp");
        th.start();
        String ln;
        while (app.isRunning()) {
            while (app.isRunning() && !reading.hasMessage()) {
                synchronized (this) {
                    try {
                        wait(ONE_SEC);
                    } catch (InterruptedException e) {
                        LOGGER.log(Level.SEVERE, "Wait interrupted", e);
                    }
                }
            }
            if (!app.isRunning()) {
                break;
            }
            ln = reading.getMessage();
            if (ln.equals(close)) {
                break;
            }
            // Pass command to application
            consoleManager.type(ln);
        }
        reading.setRunning(false);
        socket.shutdownOutput();

    }

    /** @return the consoleManager */
    public synchronized ConsoleManager getConsoleManager() {
        return consoleManager;
    }

    /** @param app the application to set */
    public synchronized void setApplication(ConsoleApplication app) {
        this.app = app;
    }

    /** This method will request the server to stop.
     * <p>
     * In most cases, this will terminate communication on every client. On some
     * cases, the */
    public void stop() {
        running = false;
        try {
            serverSocket.close();
        } catch (IOException e) {
            LOGGER.log(Level.SEVERE, "Exception in closing socket server", e); //$NON-NLS-1$
        }
        app.exit();
    }

}
